import numpy as np
from calcudoku.graph import Graph
from collections import defaultdict


def identity(x):
    return x[0]


def divide(x):
    if max(x) % min(x) == 0:
        return int(max(x) / min(x))
    else:
        return np.nan


def subtract(x):
    return np.abs(x[1] - x[0])

# Possible operations that can be done on a partition
# (min_elements, max_elements, function)
OPERATIONS = {
    'divide': (2, 2, divide),
    'multiply': (2, np.inf, np.multiply.reduce),
    'add': (2, np.inf, np.add.reduce),
    'subtract': (2, 2, subtract),
    'none': (1, 1, identity),
}

def random_board(size):
    """
    Creates a randomly generated board where there are no duplicates
    in each row/column
    """
    # Integers that control the values
    vals = list(range(1, size + 1))

    # TODO: Redo this
    board = np.zeros((size, size), dtype=int)
    for row in range(size):
        while True:
            board[row, :] = 0
            for col in range(size):
                possible = set(vals) - set(list(board[row, :]) + list(board[:, col]))

                if len(possible) == 0:
                    break
                board[row, col] = np.random.choice(list(possible))
            else:
                break
    return board.flatten()


def partition_board(size, max_partition_size,
                    initial_choice_size_factor=1,
                    merge_size_factor=0.8):
    """
    Partitions a board size into a set of groups
    """
    graph = Graph(size)

    # Repeatedly merge nodes of the graph until some conditions are satisfied
    for idx in range(int(size*size/1.3)):
        # Calculate probability of picking a node as a starting merge point
        p = np.ones(len(graph.nodes))
        for i, n in enumerate(graph.nodes):
            p[i] /= np.exp(len(n.coords))**initial_choice_size_factor
        p /= np.sum(p)
        node_idx = np.random.choice(list(range(len(graph.nodes))), p=p)

        # Create some probabilities of picking which node to merge with
        p = np.ones(len(graph.edges[node_idx]))
        for i, n in enumerate(graph.edges[node_idx]):
            p[i] /= np.exp(len(n.coords))**merge_size_factor
        p /= np.sum(p)

        merge_node = np.random.choice(graph.edges[node_idx], p=p)
        merge_idx = graph.nodes.index(merge_node)

        if len(graph.nodes[node_idx].coords) + len(merge_node.coords) > max_partition_size:
            continue

        graph.merge(node_idx, merge_idx)

    partitions = [n.coords for n in graph.nodes]

    return partitions


def possible_operations(partitions):
    all_possibles = []
    for partition in partitions:
        possibles = []

        for name, (min_size, max_size, operation) in OPERATIONS.items():
            if min_size <= len(partition) <= max_size:
                val = operation(partition)
                if ~np.isnan(val):
                    possibles.append((name, val))
        all_possibles.append(possibles)
    return all_possibles


class Calcudoku(object):
    def __init__(self):
        # Groups are dicts with 'coords', 'operation', and 'value'
        self._groups = []

        # Board is a 2d array that just stores the correct answers
        self._board = None

        # Partitions looks like [[1, 4, 5], [2], ...] which stores the connections
        # of the board coordinates
        self._partitions = None

        # Operations stores the game state, [('divide', 4), ('add', 7), ...]
        # What each partition corresponds to in the game
        self._operations = None

    @classmethod
    def generate(cls, size: int,
                 operation_p=None,
                 operation_uniformity_factor=5):
        # kwargs to control rough difficulty settings

        result = cls()

        # Start the generation by creating a random 2d board state
        # That does not have duplicates in any row or column
        result._board = random_board(size)

        # Next partition the board into a set of connected groups
        partitions = partition_board(size, max_partition_size=4)

        # Get the values corresponding to each partition location
        partition_values = []
        for p in partitions:
            partition_values.append([result._board[v] for v in p])

        possibles = possible_operations(partition_values)

        # Choose a random possible operation for each partition
        chosen = []
        chosen_operations = defaultdict(int)
        for possible in possibles:
            # Assign priorities for each possible operation
            p = np.ones(len(possible))
            for i, (op, v) in enumerate(possible):
                if operation_p:
                    p[i] *= operation_p[op]
                p[i] /= np.exp(chosen_operations[op])**operation_uniformity_factor

            p /= np.sum(p)
            # Funny things happen when taking random.choice of a list of tuples
            random_index = np.random.choice(list(range(len(possible))), p=p)
            chosen.append(possible[random_index])

            chosen_operations[possible[random_index][0]] += 1

        result._partitions = partitions
        result._operations = chosen

        return result

    @property
    def board(self):
        return self._board

    @property
    def partitions(self):
        return self._partitions

    @property
    def operations(self):
        return self._operations




class Graph(object):
    def __init__(self, size: int):
        # Start with one node for every coordinate location on the board
        self._nodes = [Node(i) for i in range(size*size)]

        # Set up the edges for each node
        self._edges = []
        for n in range(size*size):
            x = n // size
            y = n % size

            edges = [(x, y + 1), (x, y - 1), (x + 1, y), (x - 1, y)]

            to_add = []
            for e in edges:
                if e[0] >= 0 and e[0] < size and e[1] >= 0 and e[1] < size:
                    to_add.append(self._nodes[e[0] * size + e[1]])
            self._edges.append(to_add)
        pass

    def merge(self, one, two):

        # Get the unique set of edges that border the merged node
        new_edges = list(set(self._edges[one] + self._edges[two]))

        # Merge them together
        self._nodes[one].add(self._nodes[two])

        # Each node would have had the other one as an edge, so remove these
        new_edges.remove(self._nodes[one])
        new_edges.remove(self._nodes[two])

        # Update the edges on the merged node
        self._edges[one] = new_edges

        # Now we have to update the other node connections so they point to one
        # instead of two
        for edge in self._edges:
            if self._nodes[two] in edge:
                edge.remove(self._nodes[two])

                # There is a possibility where the first node was already inside edges,
                # in this case we do not need to do anything
                if self._nodes[one] not in edge:
                    edge.append(self._nodes[one])

        # Remove the second node from the structure
        del self._nodes[two]
        del self._edges[two]

    @property
    def nodes(self):
        return self._nodes

    @property
    def edges(self):
        return self._edges


class Node(object):
    def __init__(self, start):
        self._coords = [start]

    def __repr__(self):
        return 'Coords: ' + str(self.coords)

    @property
    def coords(self):
        return self._coords

    def add(self, node):
        self._coords += node.coords

    def __eq__(self, other):
        return self.coords == other.coords

    def __hash__(self):
        return hash(str(self.coords))



from setuptools import setup
import versioneer

setup(
    name='calcudoku',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=['calcudoku'],
    url='',
    license='',
    author='Daniel Zawada',
    author_email='',
    description='', install_requires=['numpy', 'matplotlib']
)



import matplotlib.pyplot as plt
from calcudoku.game import Calcudoku
import numpy as np


OPERATION_DRAW = {
    'divide': '÷',
    'multiply': 'x',
    'add': '+',
    'subtract': '-',
    'none': ''
}


def save_figure(game: Calcudoku, filename, solution=False):
    size = int(np.sqrt(len(game.board)))

    plt.figure(figsize=(size, size))

    plt.xlim(0, size)
    plt.ylim(0, size)

    plt.xticks([])
    plt.yticks([])

    for i in range(size+1):
        plt.plot([0, size], [i, i], 'k', alpha=0.25)
        plt.plot([i, i], [0, size], 'k', alpha=0.25)

    # Add in the numbers if we are printing the solution
    if solution:
        for i, number in enumerate(game.board):
            x = i // size + 0.5
            y = i % size + 0.5

            plt.text(x, y, str(number), ha='center')

    # Put in the partition lines
    for p in game.partitions:
        coords = [(i // size, i % size) for i in p]

        # For each coordinate we draw a thick box around the edges
        # Not in the partition
        for x, y in coords:
            bordering = [(x-1, y), (x+1, y), (x, y-1), (x, y+1)]
            for border in bordering:
                if border not in coords:
                    if x == border[0]:
                        diff_y = 0
                    else:
                        diff_y = 1

                    if y == border[1]:
                        diff_x = 0
                    else:
                        diff_x = 1

                    draw_x = (x + border[0])/2 + 0.5*diff_y
                    draw_y = (y + border[1])/2 + 0.5*diff_x

                    plt.plot([draw_x, draw_x + diff_x], [draw_y, draw_y + diff_y], 'k', linewidth=2)

    # Add in the operation instructions
    for p, op in zip(game.partitions, game.operations):
        # Always want to draw in the upper left, so start with finding the maximum y of the partition
        coords = [(i // size, i % size) for i in p]
        max_y = max([c[1] for c in coords])
        min_x = min([c[0] for c in coords if c[1] == max_y])

        max_y += 1 # Switch to the top of the cell

        if op[0] in OPERATION_DRAW:
            vis = str(op[1]) + OPERATION_DRAW[op[0]]
            plt.text(min_x + 0.1, max_y - 0.2, vis)

    plt.savefig(filename, bbox_inches='tight')
    pass